/* This is the utility I needed to convert my RAID arrays from the 2.2 to 
2.4 kernel on the Alpha, due to the size change of the superblock 
structure. See RH bug #38111.

Elliot Lee <sopwith@redhat.com>
*/

#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mount.h>
#include <string.h>
#include <sys/stat.h>

#define SOFF(struct_type, member)    \
    ((unsigned long) ((char*) &((struct_type*) 0)->member))

#define MINORBITS 8
#include <asm/types.h>
#include "md_p_2.4.h"
#include "md_p_2.2.h"
#define BLOCK_SIZE_BITS 10
#define BLOCK_SIZE (1<<BLOCK_SIZE_BITS)

#define MINORBITS       8
#define MINORMASK       ((1U << MINORBITS) - 1)

#define DEV_MAJOR(dev)      ((unsigned int) ((dev) >> MINORBITS))
#define DEV_MINOR(dev)      ((unsigned int) ((dev) & MINORMASK))

static void
print_disk(const char *labl, mdp_disk_t *mdd)
{
  printf("%s: number %d, device %d:%d, raid_disk %#x\n", labl,
	 mdd->number, mdd->major, mdd->minor, mdd->raid_disk);
}

int main(int argc, char *argv[])
{
  int fd, i;
  long sboff_base;
  off64_t sb_offset;
  mdp_super_2_2_t sb;
  mdp_super_t sb_out;
  char buf[32];
  int do_fixup = 0;

  if(argc > 2 && !strcmp(argv[2], "GOGOGO")) do_fixup = 1;
  fd = open(argv[1], do_fixup?O_RDWR:O_RDONLY);

  printf("2.2 mdp_super_t = %lu, 2.4 mdp_super_t = %lu\n",
	 sizeof(mdp_super_2_2_t), sizeof(mdp_super_t));
  printf("2.2 events %lu, 2.4 %lu\n",
	 SOFF(mdp_super_2_2_t, gstate_sreserved), SOFF(mdp_super_t, gstate_sreserved));
  if(ioctl(fd, BLKGETSIZE, &sboff_base))
    {
      printf("Not a block device!\n");
      return 1;
    }
  sboff_base /= 2;
  sb_offset = MD_NEW_SIZE_BLOCKS(sboff_base);
  sb_offset *= BLOCK_SIZE;
  if(lseek64(fd, sb_offset, SEEK_SET) == (off64_t)-1)
    {
      perror("lseek64 failed");
      return 1;
    }
  read(fd, &sb, MD_SB_BYTES);

  printf("%#x %#x\n", sb.md_magic, MD_SB_MAGIC);
  printf("version %d . %d . %d\n", sb.major_version, sb.minor_version,
	 sb.patch_version);

  assert(MD_22_SB_BYTES == MD_SB_BYTES);
  assert(sizeof(mdp_disk_2_2_t) == sizeof(mdp_disk_t));
  assert((MD_22_SB_PERSONALITY_WORDS - 4) == (MD_SB_PERSONALITY_WORDS - 4));
  assert((MD_22_SB_GENERIC_STATE_WORDS - 9) == (MD_SB_GENERIC_STATE_WORDS - 9));
  assert((MD_22_SB_GENERIC_CONSTANT_WORDS - 16) == (MD_SB_GENERIC_CONSTANT_WORDS - 16));
  memcpy(&sb_out, &sb, sizeof(sb_out));

  sb_out.md_magic = sb.md_magic;
  sb_out.major_version = sb.major_version;
  sb_out.patch_version = sb.patch_version;
  sb_out.gvalid_words = sb.gvalid_words;
  sb_out.set_uuid0 = sb.set_uuid0;
  sb_out.ctime = sb.ctime;
  sb_out.level = sb.level;
  sb_out.size = sb.size;
  sb_out.nr_disks = sb.nr_disks;
  sb_out.raid_disks = sb.raid_disks;
  sb_out.md_minor = sb.md_minor;
  sb_out.not_persistent = sb.not_persistent;
  sb_out.set_uuid1 = sb.set_uuid1;
  sb_out.set_uuid2 = sb.set_uuid2;
  sb_out.set_uuid3 = sb.set_uuid3;
  memcpy(sb_out.gstate_creserved, sb.gstate_creserved, sizeof(sb.gstate_creserved));
  sb_out.utime = sb.utime;
  sb_out.state = sb.state;
  sb_out.active_disks = sb.active_disks;
  sb_out.working_disks = sb.working_disks;
  sb_out.failed_disks = sb.failed_disks;
  sb_out.spare_disks = sb.spare_disks;
  sb_out.sb_csum = sb.sb_csum;
  sb_out.events_hi = (sb.events >> 32) & 0xFFFFFFFF;
  sb_out.events_lo = sb.events & 0xFFFFFFFF;
  memcpy(sb_out.gstate_sreserved, sb.gstate_sreserved, sizeof(sb.gstate_sreserved));
  sb_out.layout = sb.layout;
  sb_out.chunk_size = sb.chunk_size;
  sb_out.root_pv = sb.root_pv;
  sb_out.root_block = sb.root_block;
  memcpy(sb_out.pstate_reserved, sb.pstate_reserved, sizeof(sb.pstate_reserved));

  for(i = 0; i < sb.nr_disks; i++) {
    sprintf(buf, "%d", i);
    memcpy(&sb_out.disks[i], &sb.disks[i], sizeof(sb.disks[i]));
    print_disk(buf, &sb_out.disks[i]);
  }
  memcpy(&sb_out.this_disk, &sb.this_disk, sizeof(sb.this_disk));
  print_disk("this_disk", &sb_out.this_disk);

  if(do_fixup)
    {
      if(lseek64(fd, sb_offset, SEEK_SET) == (off64_t)-1)
	{
	  perror("lseek64 failed");
	  return 1;
	}
      write(fd, &sb_out, MD_SB_BYTES);
      fsync(fd);
    }

  return 0;
}
